// Copyright 2014 The Serviced Authors.
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// +build unit

package metrics

import (
	"encoding/json"
	"reflect"
	"sort"
	"testing"
)

// Make a MemoryUsageStats sortable by ServiceID
type SortableMemoryUsageStats []MemoryUsageStats

func (sm SortableMemoryUsageStats) Len() int {
	return len(sm)
}

func (sm SortableMemoryUsageStats) Swap(i, j int) {
	sm[i], sm[j] = sm[j], sm[i]
}

func (sm SortableMemoryUsageStats) Less(i, j int) bool {
	return sm[i].ServiceID < sm[j].ServiceID
}

func TestConvertMemoryUsage(t *testing.T) {
	testData := []byte(`
	{"clientId":"not-specified","endTime":"now","endTimeActual":1427487495,"returnset":"exact","series":true,"source":"OpenTSDB","startTime":"1m-ago","startTimeActual":1427487435,"results":[{"datapoints":[{"timestamp":1427487441,"value":2.28245504E8},{"timestamp":1427487451,"value":2.28245504E8},{"timestamp":1427487461,"value":2.28245504E8},{"timestamp":1427487471,"value":2.28245504E8},{"timestamp":1427487481,"value":2.28245504E8},{"timestamp":1427487491,"value":2.28282368E8}],"metric":"series1","tags":{"controlplane_instance_id":["0"],"controlplane_service_id":["d2b1sxt1jh7ryzuv4kszg14zv"]},"queryStatus":{"message":"","status":"SUCCESS"}},{"datapoints":[{"timestamp":1427487441,"value":9.79058688E8},{"timestamp":1427487451,"value":9.79058688E8},{"timestamp":1427487461,"value":9.78763776E8},{"timestamp":1427487471,"value":9.7875968E8},{"timestamp":1427487481,"value":9.7875968E8},{"timestamp":1427487491,"value":9.7875968E8}],"metric":"series2","tags":{"controlplane_instance_id":["0"],"controlplane_service_id":["8frik9ddntu38e3urbeimiqnf"]},"queryStatus":{"message":"","status":"SUCCESS"}}]}
	`)

	var perfdata PerformanceData
	if err := json.Unmarshal(testData, &perfdata); err != nil {
		t.Fatalf("Could not unmarshal testData: %s", err)
	}

	actual := convertMemoryUsage(&perfdata)
	expected := []MemoryUsageStats{
		{
			ServiceID:  "d2b1sxt1jh7ryzuv4kszg14zv",
			InstanceID: "0",
			Last:       int64(228282368),
			Max:        int64(228282368),
			Average:    int64(228251648),
		}, {
			ServiceID:  "8frik9ddntu38e3urbeimiqnf",
			InstanceID: "0",
			Last:       int64(978759680),
			Max:        int64(979058688),
			Average:    int64(978860032),
		},
	}

	if !reflect.DeepEqual(actual, expected) {
		t.Errorf("Expected does not equal actual: %+v", actual)
	}
}

func TestV2ConvertMemoryUsage(t *testing.T) {
	testData := make(map[string][]byte)
	testData["last"] = []byte(`
	{ "series" : [ { "datapoints" : [ [ 1453835068, 3.2763904E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1u5drlen4es4nxrss41o68xa" } }, { "datapoints" : [ [ 1453835068, 3.9354368E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4sopamwpjthi61cxymjzugnyt" } }, { "datapoints" : [ [ 1453835068, 3.6716544E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6r9mtb8cgcjjl9k0i3fv3pq9b" } }, { "datapoints" : [ [ 1453835068, 8.16472064E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ec1f7icsl924plv8a5ds8d0pc" } }, { "datapoints" : [ [ 1453835068, 4.50469888E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bcn05pjmagb478zgguqbplqr7" } }, { "datapoints" : [ [ 1453835068, 9.9463168E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "760h3bj64lri3mrha4vyo0mj2" } }, { "datapoints" : [ [ 1453835068, 4.16366592E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "5mmmwxdctw0lym6jg9zt7z4bd" } }, { "datapoints" : [ [ 1453835068, 2.5247744E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7mzn9qp8gugal14ivtu5uggdf" } }, { "datapoints" : [ [ 1453835068, 1.65466112E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7ug2jnoyjjgmfi0nx1ad5vwy2" } }, { "datapoints" : [ [ 1453835068, 4.1558016E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3krktjw1z4fow86gmym0uqq0d" } }, { "datapoints" : [ [ 1453835068, 4.93490176E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bpajma004iirdb512rstb3mdq" } }, { "datapoints" : [ [ 1453835068, 4.4265472E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bbm5cp87tfwltopv1xb1sa5x6" } }, { "datapoints" : [ [ 1453835068, 3.6532224E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "848v3up41cybjms3v20ibdm7o" } }, { "datapoints" : [ [ 1453835068, 2.1909504E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "90qbe5y3h9r2012pvjcs9c65y" } }, { "datapoints" : [ [ 1453835068, 1.39739136E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4no3uenok47urez0e6dkucy22" } }, { "datapoints" : [ [ 1453835068, 1.37183232E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "68xw1lsloq2iomj94ckhc7zil" } }, { "datapoints" : [ [ 1453835068, 1.3955072E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "d5h1sr4oxlulm6jzvprscrs83" } }, { "datapoints" : [ [ 1453835068, 3.66800896E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "8b97m6ikb4fpmgwytxro3lgyw" } }, { "datapoints" : [ [ 1453835068, 1.35929856E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3bdwp70aqnjcluorkahp0sjpe" } }, { "datapoints" : [ [ 1453835068, 1.37109504E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "e3f7igy5lss9gpgspwcd04od" } }, { "datapoints" : [ [ 1453835068, 1.3236224E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1y37j22er1o64ri70asc28v9" } }, { "datapoints" : [ [ 1453835068, 1.32927488E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "dckztpu9kod72t9dw6jnkcq3e" } }, { "datapoints" : [ [ 1453835068, 3.534848E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "92q196piagp8u8106obz4gpxd" } }, { "datapoints" : [ [ 1453835068, 1.33173248E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3b35h6gp7md4fk4cgjqme5xdp" } }, { "datapoints" : [ [ 1453835068, 2.52145664E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "31y8o1ceiits9qlqmyelyr6x7" } }, { "datapoints" : [ [ 1453835068, 1.34545408E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "agivpghlm70ki06xf4ysv06ek" } }, { "datapoints" : [ [ 1453835068, 3.72969472E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "asf5d2etnkqle4lxltentgbjs" } }, { "datapoints" : [ [ 1453835068, 1.49532672E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "r382ljbz4y594stdrqro042u" } }, { "datapoints" : [ [ 1453835068, 1.34213632E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ct9sze964cjf3z1jdja2byezf" } }, { "datapoints" : [ [ 1453835068, 1.92499712E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6gvzpnjs6yl2ujhvzzz2g0ypz" } } ], "statuses" : [ { "message" : "", "status" : "SUCCESS" } ] }
	`)
	testData["max"] = []byte(`
	{ "series" : [ { "datapoints" : [ [ 1453766400, 4.84724736E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1u5drlen4es4nxrss41o68xa" } }, { "datapoints" : [ [ 1453766400, 4.0443904E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4sopamwpjthi61cxymjzugnyt" } }, { "datapoints" : [ [ 1453766400, 3.6753408E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6r9mtb8cgcjjl9k0i3fv3pq9b" } }, { "datapoints" : [ [ 1453766400, 8.35043328E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ec1f7icsl924plv8a5ds8d0pc" } }, { "datapoints" : [ [ 1453766400, 4.91388928E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bcn05pjmagb478zgguqbplqr7" } }, { "datapoints" : [ [ 1453766400, 1.47070976E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "760h3bj64lri3mrha4vyo0mj2" } }, { "datapoints" : [ [ 1453766400, 4.17456128E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "5mmmwxdctw0lym6jg9zt7z4bd" } }, { "datapoints" : [ [ 1453766400, 2.67264E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7mzn9qp8gugal14ivtu5uggdf" } }, { "datapoints" : [ [ 1453766400, 1.6562176E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7ug2jnoyjjgmfi0nx1ad5vwy2" } }, { "datapoints" : [ [ 1453766400, 4.3593728E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3krktjw1z4fow86gmym0uqq0d" } }, { "datapoints" : [ [ 1453766400, 5.28248832E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bpajma004iirdb512rstb3mdq" } }, { "datapoints" : [ [ 1453766400, 4.7742976E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bbm5cp87tfwltopv1xb1sa5x6" } }, { "datapoints" : [ [ 1453766400, 2.50617856E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "848v3up41cybjms3v20ibdm7o" } }, { "datapoints" : [ [ 1453766400, 2.1917696E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "90qbe5y3h9r2012pvjcs9c65y" } }, { "datapoints" : [ [ 1453766400, 1.39804672E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4no3uenok47urez0e6dkucy22" } }, { "datapoints" : [ [ 1453766400, 1.37408512E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "68xw1lsloq2iomj94ckhc7zil" } }, { "datapoints" : [ [ 1453766400, 1.395712E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "d5h1sr4oxlulm6jzvprscrs83" } }, { "datapoints" : [ [ 1453766400, 3.76475648E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "8b97m6ikb4fpmgwytxro3lgyw" } }, { "datapoints" : [ [ 1453766400, 1.375232E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3bdwp70aqnjcluorkahp0sjpe" } }, { "datapoints" : [ [ 1453766400, 1.37310208E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "e3f7igy5lss9gpgspwcd04od" } }, { "datapoints" : [ [ 1453766400, 1.32493312E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1y37j22er1o64ri70asc28v9" } }, { "datapoints" : [ [ 1453766400, 1.3305856E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "dckztpu9kod72t9dw6jnkcq3e" } }, { "datapoints" : [ [ 1453766400, 3.5385344E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "92q196piagp8u8106obz4gpxd" } }, { "datapoints" : [ [ 1453766400, 1.4065664E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3b35h6gp7md4fk4cgjqme5xdp" } }, { "datapoints" : [ [ 1453766400, 2.5257984E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "31y8o1ceiits9qlqmyelyr6x7" } }, { "datapoints" : [ [ 1453766400, 1.40689408E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "agivpghlm70ki06xf4ysv06ek" } }, { "datapoints" : [ [ 1453766400, 3.73030912E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "asf5d2etnkqle4lxltentgbjs" } }, { "datapoints" : [ [ 1453766400, 1.49536768E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "r382ljbz4y594stdrqro042u" } }, { "datapoints" : [ [ 1453766400, 1.35806976E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ct9sze964cjf3z1jdja2byezf" } }, { "datapoints" : [ [ 1453766400, 1.93114112E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6gvzpnjs6yl2ujhvzzz2g0ypz" } } ], "statuses" : [ { "message" : "", "status" : "SUCCESS" } ] }
	`)
	testData["avg"] = []byte(`
	{ "series" : [ { "datapoints" : [ [ 1453766400, 3.0809602071011233E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1u5drlen4es4nxrss41o68xa" } }, { "datapoints" : [ [ 1453766400, 1.5352875721348315E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4sopamwpjthi61cxymjzugnyt" } }, { "datapoints" : [ [ 1453766400, 3.391377723076923E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6r9mtb8cgcjjl9k0i3fv3pq9b" } }, { "datapoints" : [ [ 1453766400, 8.217153517466063E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ec1f7icsl924plv8a5ds8d0pc" } }, { "datapoints" : [ [ 1453766400, 4.4049607239819E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bcn05pjmagb478zgguqbplqr7" } }, { "datapoints" : [ [ 1453766400, 9.879526052488688E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "760h3bj64lri3mrha4vyo0mj2" } }, { "datapoints" : [ [ 1453766400, 3.965997511111111E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "5mmmwxdctw0lym6jg9zt7z4bd" } }, { "datapoints" : [ [ 1453766400, 2.3452507138321996E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7mzn9qp8gugal14ivtu5uggdf" } }, { "datapoints" : [ [ 1453766400, 1.4289896315646258E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7ug2jnoyjjgmfi0nx1ad5vwy2" } }, { "datapoints" : [ [ 1453766400, 3.8384981333333336E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3krktjw1z4fow86gmym0uqq0d" } }, { "datapoints" : [ [ 1453766400, 4.1184256E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bpajma004iirdb512rstb3mdq" } }, { "datapoints" : [ [ 1453766400, 4.1428519384615384E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bbm5cp87tfwltopv1xb1sa5x6" } }, { "datapoints" : [ [ 1453766400, 5.1902621538461536E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "848v3up41cybjms3v20ibdm7o" } }, { "datapoints" : [ [ 1453766400, 2.0540245333333332E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "90qbe5y3h9r2012pvjcs9c65y" } }, { "datapoints" : [ [ 1453766400, 1.286760448E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4no3uenok47urez0e6dkucy22" } }, { "datapoints" : [ [ 1453766400, 1.3061885305263157E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "68xw1lsloq2iomj94ckhc7zil" } }, { "datapoints" : [ [ 1453766400, 1.2429010189473684E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "d5h1sr4oxlulm6jzvprscrs83" } }, { "datapoints" : [ [ 1453766400, 2.7670793035294116E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "8b97m6ikb4fpmgwytxro3lgyw" } }, { "datapoints" : [ [ 1453766400, 1.2192394541176471E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3bdwp70aqnjcluorkahp0sjpe" } }, { "datapoints" : [ [ 1453766400, 1.2655676235294117E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "e3f7igy5lss9gpgspwcd04od" } }, { "datapoints" : [ [ 1453766400, 1.2112787576470588E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1y37j22er1o64ri70asc28v9" } }, { "datapoints" : [ [ 1453766400, 1.2106233976470588E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "dckztpu9kod72t9dw6jnkcq3e" } }, { "datapoints" : [ [ 1453766400, 3.333421176470588E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "92q196piagp8u8106obz4gpxd" } }, { "datapoints" : [ [ 1453766400, 1.1327945788235295E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3b35h6gp7md4fk4cgjqme5xdp" } }, { "datapoints" : [ [ 1453766400, 2.182250014117647E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "31y8o1ceiits9qlqmyelyr6x7" } }, { "datapoints" : [ [ 1453766400, 1.13495296E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "agivpghlm70ki06xf4ysv06ek" } }, { "datapoints" : [ [ 1453766400, 2.9701504E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "asf5d2etnkqle4lxltentgbjs" } }, { "datapoints" : [ [ 1453766400, 1.27629568E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "r382ljbz4y594stdrqro042u" } }, { "datapoints" : [ [ 1453766400, 1.05186048E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ct9sze964cjf3z1jdja2byezf" } }, { "datapoints" : [ [ 1453766400, 1.8303488E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6gvzpnjs6yl2ujhvzzz2g0ypz" } } ], "statuses" : [ { "message" : "", "status" : "SUCCESS" } ] }
	`)

	perfMap := make(map[string]*V2PerformanceData)
	for _, agg := range []string{"max", "avg", "last"} {
		var perfdata V2PerformanceData
		if err := json.Unmarshal(testData[agg], &perfdata); err != nil {
			t.Fatalf("Could not unmarshal testData: %s", err)
		}
		perfMap[agg] = &perfdata
	}

	actual := convertV2MemoryUsage(perfMap)
	expected := []MemoryUsageStats{
		MemoryUsageStats{HostID: "007f0101", ServiceID: "5mmmwxdctw0lym6jg9zt7z4bd", InstanceID: "0", Last: 416366592, Max: 417456128, Average: 396599751},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "7mzn9qp8gugal14ivtu5uggdf", InstanceID: "0", Last: 25247744, Max: 26726400, Average: 23452507},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "ct9sze964cjf3z1jdja2byezf", InstanceID: "0", Last: 134213632, Max: 135806976, Average: 105186048},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "760h3bj64lri3mrha4vyo0mj2", InstanceID: "0", Last: 99463168, Max: 147070976, Average: 98795260},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "bbm5cp87tfwltopv1xb1sa5x6", InstanceID: "0", Last: 44265472, Max: 47742976, Average: 41428519},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "agivpghlm70ki06xf4ysv06ek", InstanceID: "0", Last: 134545408, Max: 140689408, Average: 113495296},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "r382ljbz4y594stdrqro042u", InstanceID: "0", Last: 149532672, Max: 149536768, Average: 127629568},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "3krktjw1z4fow86gmym0uqq0d", InstanceID: "0", Last: 41558016, Max: 43593728, Average: 38384981},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "3bdwp70aqnjcluorkahp0sjpe", InstanceID: "0", Last: 135929856, Max: 137523200, Average: 121923945},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "f1u5drlen4es4nxrss41o68xa", InstanceID: "0", Last: 327639040, Max: 484724736, Average: 308096020},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "4sopamwpjthi61cxymjzugnyt", InstanceID: "0", Last: 39354368, Max: 40443904, Average: 15352875},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "4no3uenok47urez0e6dkucy22", InstanceID: "0", Last: 139739136, Max: 139804672, Average: 128676044},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "68xw1lsloq2iomj94ckhc7zil", InstanceID: "0", Last: 137183232, Max: 137408512, Average: 130618853},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "90qbe5y3h9r2012pvjcs9c65y", InstanceID: "0", Last: 21909504, Max: 21917696, Average: 20540245},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "f1y37j22er1o64ri70asc28v9", InstanceID: "0", Last: 132362240, Max: 132493312, Average: 121127875},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "dckztpu9kod72t9dw6jnkcq3e", InstanceID: "0", Last: 132927488, Max: 133058560, Average: 121062339},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "asf5d2etnkqle4lxltentgbjs", InstanceID: "0", Last: 372969472, Max: 373030912, Average: 297015040},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "31y8o1ceiits9qlqmyelyr6x7", InstanceID: "0", Last: 252145664, Max: 252579840, Average: 218225001},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "d5h1sr4oxlulm6jzvprscrs83", InstanceID: "0", Last: 139550720, Max: 139571200, Average: 124290101},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "e3f7igy5lss9gpgspwcd04od", InstanceID: "0", Last: 137109504, Max: 137310208, Average: 126556762},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "92q196piagp8u8106obz4gpxd", InstanceID: "0", Last: 35348480, Max: 35385344, Average: 33334211},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "3b35h6gp7md4fk4cgjqme5xdp", InstanceID: "0", Last: 133173248, Max: 140656640, Average: 113279457},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "ec1f7icsl924plv8a5ds8d0pc", InstanceID: "0", Last: 816472064, Max: 835043328, Average: 821715351},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "bpajma004iirdb512rstb3mdq", InstanceID: "0", Last: 493490176, Max: 528248832, Average: 411842560},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "8b97m6ikb4fpmgwytxro3lgyw", InstanceID: "0", Last: 366800896, Max: 376475648, Average: 276707930},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "6gvzpnjs6yl2ujhvzzz2g0ypz", InstanceID: "0", Last: 192499712, Max: 193114112, Average: 183034880},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "6r9mtb8cgcjjl9k0i3fv3pq9b", InstanceID: "0", Last: 36716544, Max: 36753408, Average: 33913777},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "bcn05pjmagb478zgguqbplqr7", InstanceID: "0", Last: 450469888, Max: 491388928, Average: 440496072},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "7ug2jnoyjjgmfi0nx1ad5vwy2", InstanceID: "0", Last: 165466112, Max: 165621760, Average: 142898963},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "848v3up41cybjms3v20ibdm7o", InstanceID: "0", Last: 36532224, Max: 250617856, Average: 51902621},
	}

	sort.Sort(SortableMemoryUsageStats(actual))
	sort.Sort(SortableMemoryUsageStats(expected))

	if !reflect.DeepEqual(actual, expected) {
		t.Errorf("Expected does not equal actual: %+v", actual)
	}

}

func TestBadV2ConvertMemoryUsage(t *testing.T) {
	testData := make(map[string][]byte)
	// The first series' datapoints is empty
	testData["last"] = []byte(`
	{ "series" : [ { "datapoints" : [ ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1u5drlen4es4nxrss41o68xa" } }, { "datapoints" : [ [ 1453835068, 3.9354368E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4sopamwpjthi61cxymjzugnyt" } }, { "datapoints" : [ [ 1453835068, 3.6716544E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6r9mtb8cgcjjl9k0i3fv3pq9b" } }, { "datapoints" : [ [ 1453835068, 8.16472064E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ec1f7icsl924plv8a5ds8d0pc" } }, { "datapoints" : [ [ 1453835068, 4.50469888E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bcn05pjmagb478zgguqbplqr7" } }, { "datapoints" : [ [ 1453835068, 9.9463168E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "760h3bj64lri3mrha4vyo0mj2" } }, { "datapoints" : [ [ 1453835068, 4.16366592E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "5mmmwxdctw0lym6jg9zt7z4bd" } }, { "datapoints" : [ [ 1453835068, 2.5247744E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7mzn9qp8gugal14ivtu5uggdf" } }, { "datapoints" : [ [ 1453835068, 1.65466112E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7ug2jnoyjjgmfi0nx1ad5vwy2" } }, { "datapoints" : [ [ 1453835068, 4.1558016E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3krktjw1z4fow86gmym0uqq0d" } }, { "datapoints" : [ [ 1453835068, 4.93490176E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bpajma004iirdb512rstb3mdq" } }, { "datapoints" : [ [ 1453835068, 4.4265472E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bbm5cp87tfwltopv1xb1sa5x6" } }, { "datapoints" : [ [ 1453835068, 3.6532224E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "848v3up41cybjms3v20ibdm7o" } }, { "datapoints" : [ [ 1453835068, 2.1909504E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "90qbe5y3h9r2012pvjcs9c65y" } }, { "datapoints" : [ [ 1453835068, 1.39739136E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4no3uenok47urez0e6dkucy22" } }, { "datapoints" : [ [ 1453835068, 1.37183232E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "68xw1lsloq2iomj94ckhc7zil" } }, { "datapoints" : [ [ 1453835068, 1.3955072E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "d5h1sr4oxlulm6jzvprscrs83" } }, { "datapoints" : [ [ 1453835068, 3.66800896E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "8b97m6ikb4fpmgwytxro3lgyw" } }, { "datapoints" : [ [ 1453835068, 1.35929856E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3bdwp70aqnjcluorkahp0sjpe" } }, { "datapoints" : [ [ 1453835068, 1.37109504E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "e3f7igy5lss9gpgspwcd04od" } }, { "datapoints" : [ [ 1453835068, 1.3236224E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1y37j22er1o64ri70asc28v9" } }, { "datapoints" : [ [ 1453835068, 1.32927488E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "dckztpu9kod72t9dw6jnkcq3e" } }, { "datapoints" : [ [ 1453835068, 3.534848E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "92q196piagp8u8106obz4gpxd" } }, { "datapoints" : [ [ 1453835068, 1.33173248E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3b35h6gp7md4fk4cgjqme5xdp" } }, { "datapoints" : [ [ 1453835068, 2.52145664E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "31y8o1ceiits9qlqmyelyr6x7" } }, { "datapoints" : [ [ 1453835068, 1.34545408E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "agivpghlm70ki06xf4ysv06ek" } }, { "datapoints" : [ [ 1453835068, 3.72969472E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "asf5d2etnkqle4lxltentgbjs" } }, { "datapoints" : [ [ 1453835068, 1.49532672E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "r382ljbz4y594stdrqro042u" } }, { "datapoints" : [ [ 1453835068, 1.34213632E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ct9sze964cjf3z1jdja2byezf" } }, { "datapoints" : [ [ 1453835068, 1.92499712E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6gvzpnjs6yl2ujhvzzz2g0ypz" } } ], "statuses" : [ { "message" : "", "status" : "SUCCESS" } ] }
	`)
	testData["max"] = []byte(`
	{ "series" : [ { "datapoints" : [ [ 1453766400, 4.84724736E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1u5drlen4es4nxrss41o68xa" } }, { "datapoints" : [ [ 1453766400, 4.0443904E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4sopamwpjthi61cxymjzugnyt" } }, { "datapoints" : [ [ 1453766400, 3.6753408E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6r9mtb8cgcjjl9k0i3fv3pq9b" } }, { "datapoints" : [ [ 1453766400, 8.35043328E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ec1f7icsl924plv8a5ds8d0pc" } }, { "datapoints" : [ [ 1453766400, 4.91388928E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bcn05pjmagb478zgguqbplqr7" } }, { "datapoints" : [ [ 1453766400, 1.47070976E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "760h3bj64lri3mrha4vyo0mj2" } }, { "datapoints" : [ [ 1453766400, 4.17456128E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "5mmmwxdctw0lym6jg9zt7z4bd" } }, { "datapoints" : [ [ 1453766400, 2.67264E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7mzn9qp8gugal14ivtu5uggdf" } }, { "datapoints" : [ [ 1453766400, 1.6562176E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7ug2jnoyjjgmfi0nx1ad5vwy2" } }, { "datapoints" : [ [ 1453766400, 4.3593728E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3krktjw1z4fow86gmym0uqq0d" } }, { "datapoints" : [ [ 1453766400, 5.28248832E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bpajma004iirdb512rstb3mdq" } }, { "datapoints" : [ [ 1453766400, 4.7742976E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bbm5cp87tfwltopv1xb1sa5x6" } }, { "datapoints" : [ [ 1453766400, 2.50617856E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "848v3up41cybjms3v20ibdm7o" } }, { "datapoints" : [ [ 1453766400, 2.1917696E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "90qbe5y3h9r2012pvjcs9c65y" } }, { "datapoints" : [ [ 1453766400, 1.39804672E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4no3uenok47urez0e6dkucy22" } }, { "datapoints" : [ [ 1453766400, 1.37408512E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "68xw1lsloq2iomj94ckhc7zil" } }, { "datapoints" : [ [ 1453766400, 1.395712E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "d5h1sr4oxlulm6jzvprscrs83" } }, { "datapoints" : [ [ 1453766400, 3.76475648E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "8b97m6ikb4fpmgwytxro3lgyw" } }, { "datapoints" : [ [ 1453766400, 1.375232E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3bdwp70aqnjcluorkahp0sjpe" } }, { "datapoints" : [ [ 1453766400, 1.37310208E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "e3f7igy5lss9gpgspwcd04od" } }, { "datapoints" : [ [ 1453766400, 1.32493312E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1y37j22er1o64ri70asc28v9" } }, { "datapoints" : [ [ 1453766400, 1.3305856E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "dckztpu9kod72t9dw6jnkcq3e" } }, { "datapoints" : [ [ 1453766400, 3.5385344E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "92q196piagp8u8106obz4gpxd" } }, { "datapoints" : [ [ 1453766400, 1.4065664E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3b35h6gp7md4fk4cgjqme5xdp" } }, { "datapoints" : [ [ 1453766400, 2.5257984E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "31y8o1ceiits9qlqmyelyr6x7" } }, { "datapoints" : [ [ 1453766400, 1.40689408E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "agivpghlm70ki06xf4ysv06ek" } }, { "datapoints" : [ [ 1453766400, 3.73030912E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "asf5d2etnkqle4lxltentgbjs" } }, { "datapoints" : [ [ 1453766400, 1.49536768E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "r382ljbz4y594stdrqro042u" } }, { "datapoints" : [ [ 1453766400, 1.35806976E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ct9sze964cjf3z1jdja2byezf" } }, { "datapoints" : [ [ 1453766400, 1.93114112E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6gvzpnjs6yl2ujhvzzz2g0ypz" } } ], "statuses" : [ { "message" : "", "status" : "SUCCESS" } ] }
	`)
	testData["avg"] = []byte(`
	{ "series" : [ { "datapoints" : [ [ 1453766400, 3.0809602071011233E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1u5drlen4es4nxrss41o68xa" } }, { "datapoints" : [ [ 1453766400, 1.5352875721348315E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4sopamwpjthi61cxymjzugnyt" } }, { "datapoints" : [ [ 1453766400, 3.391377723076923E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6r9mtb8cgcjjl9k0i3fv3pq9b" } }, { "datapoints" : [ [ 1453766400, 8.217153517466063E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ec1f7icsl924plv8a5ds8d0pc" } }, { "datapoints" : [ [ 1453766400, 4.4049607239819E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bcn05pjmagb478zgguqbplqr7" } }, { "datapoints" : [ [ 1453766400, 9.879526052488688E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "760h3bj64lri3mrha4vyo0mj2" } }, { "datapoints" : [ [ 1453766400, 3.965997511111111E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "5mmmwxdctw0lym6jg9zt7z4bd" } }, { "datapoints" : [ [ 1453766400, 2.3452507138321996E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7mzn9qp8gugal14ivtu5uggdf" } }, { "datapoints" : [ [ 1453766400, 1.4289896315646258E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "7ug2jnoyjjgmfi0nx1ad5vwy2" } }, { "datapoints" : [ [ 1453766400, 3.8384981333333336E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3krktjw1z4fow86gmym0uqq0d" } }, { "datapoints" : [ [ 1453766400, 4.1184256E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bpajma004iirdb512rstb3mdq" } }, { "datapoints" : [ [ 1453766400, 4.1428519384615384E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "bbm5cp87tfwltopv1xb1sa5x6" } }, { "datapoints" : [ [ 1453766400, 5.1902621538461536E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "848v3up41cybjms3v20ibdm7o" } }, { "datapoints" : [ [ 1453766400, 2.0540245333333332E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "90qbe5y3h9r2012pvjcs9c65y" } }, { "datapoints" : [ [ 1453766400, 1.286760448E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "4no3uenok47urez0e6dkucy22" } }, { "datapoints" : [ [ 1453766400, 1.3061885305263157E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "68xw1lsloq2iomj94ckhc7zil" } }, { "datapoints" : [ [ 1453766400, 1.2429010189473684E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "d5h1sr4oxlulm6jzvprscrs83" } }, { "datapoints" : [ [ 1453766400, 2.7670793035294116E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "8b97m6ikb4fpmgwytxro3lgyw" } }, { "datapoints" : [ [ 1453766400, 1.2192394541176471E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3bdwp70aqnjcluorkahp0sjpe" } }, { "datapoints" : [ [ 1453766400, 1.2655676235294117E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "e3f7igy5lss9gpgspwcd04od" } }, { "datapoints" : [ [ 1453766400, 1.2112787576470588E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "f1y37j22er1o64ri70asc28v9" } }, { "datapoints" : [ [ 1453766400, 1.2106233976470588E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "dckztpu9kod72t9dw6jnkcq3e" } }, { "datapoints" : [ [ 1453766400, 3.333421176470588E7 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "92q196piagp8u8106obz4gpxd" } }, { "datapoints" : [ [ 1453766400, 1.1327945788235295E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "3b35h6gp7md4fk4cgjqme5xdp" } }, { "datapoints" : [ [ 1453766400, 2.182250014117647E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "31y8o1ceiits9qlqmyelyr6x7" } }, { "datapoints" : [ [ 1453766400, 1.13495296E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "agivpghlm70ki06xf4ysv06ek" } }, { "datapoints" : [ [ 1453766400, 2.9701504E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "asf5d2etnkqle4lxltentgbjs" } }, { "datapoints" : [ [ 1453766400, 1.27629568E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "r382ljbz4y594stdrqro042u" } }, { "datapoints" : [ [ 1453766400, 1.05186048E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "ct9sze964cjf3z1jdja2byezf" } }, { "datapoints" : [ [ 1453766400, 1.8303488E8 ] ], "metric" : "cgroup.memory.totalrss", "tags" : { "controlplane_instance_id" : "0", "controlplane_host_id" : "007f0101", "controlplane_service_id" : "6gvzpnjs6yl2ujhvzzz2g0ypz" } } ], "statuses" : [ { "message" : "", "status" : "SUCCESS" } ] }
	`)

	perfMap := make(map[string]*V2PerformanceData)
	for _, agg := range []string{"max", "avg", "last"} {
		var perfdata V2PerformanceData
		if err := json.Unmarshal(testData[agg], &perfdata); err != nil {
			t.Fatalf("Could not unmarshal testData: %s", err)
		}
		perfMap[agg] = &perfdata
	}

	actual := convertV2MemoryUsage(perfMap)
	expected := []MemoryUsageStats{
		MemoryUsageStats{HostID: "007f0101", ServiceID: "5mmmwxdctw0lym6jg9zt7z4bd", InstanceID: "0", Last: 416366592, Max: 417456128, Average: 396599751},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "7mzn9qp8gugal14ivtu5uggdf", InstanceID: "0", Last: 25247744, Max: 26726400, Average: 23452507},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "ct9sze964cjf3z1jdja2byezf", InstanceID: "0", Last: 134213632, Max: 135806976, Average: 105186048},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "760h3bj64lri3mrha4vyo0mj2", InstanceID: "0", Last: 99463168, Max: 147070976, Average: 98795260},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "bbm5cp87tfwltopv1xb1sa5x6", InstanceID: "0", Last: 44265472, Max: 47742976, Average: 41428519},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "agivpghlm70ki06xf4ysv06ek", InstanceID: "0", Last: 134545408, Max: 140689408, Average: 113495296},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "r382ljbz4y594stdrqro042u", InstanceID: "0", Last: 149532672, Max: 149536768, Average: 127629568},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "3krktjw1z4fow86gmym0uqq0d", InstanceID: "0", Last: 41558016, Max: 43593728, Average: 38384981},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "3bdwp70aqnjcluorkahp0sjpe", InstanceID: "0", Last: 135929856, Max: 137523200, Average: 121923945},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "f1u5drlen4es4nxrss41o68xa", InstanceID: "0", Last: 0, Max: 484724736, Average: 308096020},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "4sopamwpjthi61cxymjzugnyt", InstanceID: "0", Last: 39354368, Max: 40443904, Average: 15352875},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "4no3uenok47urez0e6dkucy22", InstanceID: "0", Last: 139739136, Max: 139804672, Average: 128676044},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "68xw1lsloq2iomj94ckhc7zil", InstanceID: "0", Last: 137183232, Max: 137408512, Average: 130618853},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "90qbe5y3h9r2012pvjcs9c65y", InstanceID: "0", Last: 21909504, Max: 21917696, Average: 20540245},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "f1y37j22er1o64ri70asc28v9", InstanceID: "0", Last: 132362240, Max: 132493312, Average: 121127875},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "dckztpu9kod72t9dw6jnkcq3e", InstanceID: "0", Last: 132927488, Max: 133058560, Average: 121062339},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "asf5d2etnkqle4lxltentgbjs", InstanceID: "0", Last: 372969472, Max: 373030912, Average: 297015040},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "31y8o1ceiits9qlqmyelyr6x7", InstanceID: "0", Last: 252145664, Max: 252579840, Average: 218225001},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "d5h1sr4oxlulm6jzvprscrs83", InstanceID: "0", Last: 139550720, Max: 139571200, Average: 124290101},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "e3f7igy5lss9gpgspwcd04od", InstanceID: "0", Last: 137109504, Max: 137310208, Average: 126556762},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "92q196piagp8u8106obz4gpxd", InstanceID: "0", Last: 35348480, Max: 35385344, Average: 33334211},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "3b35h6gp7md4fk4cgjqme5xdp", InstanceID: "0", Last: 133173248, Max: 140656640, Average: 113279457},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "ec1f7icsl924plv8a5ds8d0pc", InstanceID: "0", Last: 816472064, Max: 835043328, Average: 821715351},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "bpajma004iirdb512rstb3mdq", InstanceID: "0", Last: 493490176, Max: 528248832, Average: 411842560},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "8b97m6ikb4fpmgwytxro3lgyw", InstanceID: "0", Last: 366800896, Max: 376475648, Average: 276707930},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "6gvzpnjs6yl2ujhvzzz2g0ypz", InstanceID: "0", Last: 192499712, Max: 193114112, Average: 183034880},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "6r9mtb8cgcjjl9k0i3fv3pq9b", InstanceID: "0", Last: 36716544, Max: 36753408, Average: 33913777},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "bcn05pjmagb478zgguqbplqr7", InstanceID: "0", Last: 450469888, Max: 491388928, Average: 440496072},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "7ug2jnoyjjgmfi0nx1ad5vwy2", InstanceID: "0", Last: 165466112, Max: 165621760, Average: 142898963},
		MemoryUsageStats{HostID: "007f0101", ServiceID: "848v3up41cybjms3v20ibdm7o", InstanceID: "0", Last: 36532224, Max: 250617856, Average: 51902621},
	}

	sort.Sort(SortableMemoryUsageStats(actual))
	sort.Sort(SortableMemoryUsageStats(expected))

	if !reflect.DeepEqual(actual, expected) {
		t.Errorf("Expected does not equal actual: %+v", actual)
	}

}
